#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""MySQL binlog to Google Pub/Sub entry point"""
from __future__ import print_function
import logging
import logging.config
import argparse
import sys
import os


__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


def _setup_arg_parser(argv):
    """ Parse arguments from command line
    Args:
        argv list: by default it's command line arguments

    Returns:
        argparse.Namespace: parsed argument
    """
    parser = argparse.ArgumentParser(
        description='MySQL binlog to Google Cloud Pub/Sub')
    parser.add_argument('conf',
                        help='configuration file for publishing')
    parser.add_argument('--loglevel', '-l', default=None,
                        help='log level for root')

    if os.path.isfile('logging.ini'):
        _log_file = 'logging.ini'
    else:
        _log_file = None

    parser.add_argument('--logconf', default=_log_file,
                        help='INI file log configuration')
    args = parser.parse_args(argv)
    return args


def main(argv):
    """ MySQL binlog to Google Pub/Sub entry point
    Args:
        argv (list): list of command line arguments
    """
    args = _setup_arg_parser(argv)
    conf_file = args.conf

    if conf_file:
        os.environ['BINLOG2GPUBSUB_CONF_FILE'] = conf_file

    if args.logconf:
        logging.config.fileConfig(args.logconf, disable_existing_loggers=False)
    else:
        logging.basicConfig()

    if args.loglevel:
        logging.root.setLevel(logging.getLevelName(args.loglevel.upper()))

    import mysqlbinlog2gpubsub
    mysqlbinlog2gpubsub.start_publishing()


def entry_point():
    """Zero-argument entry point for use with setuptools/distribute."""
    raise SystemExit(main(sys.argv[1:]))


if __name__ == '__main__':
    entry_point()


# -*- coding: utf-8 -*-
""" This module will convert a BinlogRowEvent to a dict data and meta data """
import logging
from pymysqlreplication import row_event
from . import config

__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


def _get_event_action(event):
    if isinstance(event, row_event.WriteRowsEvent):
        return 'insert'
    if isinstance(event, row_event.UpdateRowsEvent):
        return 'update'
    if isinstance(event, row_event.DeleteRowsEvent):
        return 'delete'
    assert False, "Can not detect event action"


def _get_updated_values(before_values, after_values):
    """ Get updated values from 2 dict of values
    :param dict before_values: values before update
    :param dict after_values: values after update
    :return: a diff dict with key is field key, value is tuple of
    (before_value, after_value)
    :rtype: dict
    """
    assert before_values.keys() == after_values.keys()
    return dict([(k, [before_values[k], after_values[k]])
                 for k in before_values.keys()
                 if before_values[k] != after_values[k]])


class TableAdapter(object):
    def __init__(self, schema, table):
        self.schema = schema
        self.table = table
        self.prefix = config.TABLE_PREFIX.get(table, '')
        self.prefix_length = len(self.prefix)
        self.ignored_update_fields = \
            config.TABLE_IGNORED_UPDATE_FIELDS.get(table, [])

        if self.prefix_length:
            self.remove_prefix_for_list = self._do_remove_prefix_for_list
            self.remove_prefix_for_dict = self._do_remove_prefix_for_dict
        else:
            self.remove_prefix_for_list = self._do_nothing
            self.remove_prefix_for_dict = self._do_nothing
        if self.ignored_update_fields:
            self.remove_ignored_fields = self._do_remove_ignored_fields
        else:
            self.remove_ignored_fields = self._do_nothing
        self.key_fields = config.TABLE_PRIMARY_KEYS[self.table]

    def _do_nothing(self, val):
        return val

    def _do_remove_prefix_for_list(self, vals):
        return map(
            lambda v: v[self.prefix_length:] if v.startswith(self.prefix) else v,
            vals)

    def _do_remove_prefix_for_dict(self, data):
        return {
            k[self.prefix_length:] if k.startswith(self.prefix) else k: v
            for k, v in data.items()
        }

    def _do_remove_ignored_fields(self, data):
        return dict([(k, v)
                     for k, v in data.items()
                     if k not in self.ignored_update_fields])

    def refine_insert_row(self, row):
        return {
            'values': self.remove_prefix_for_dict(row['values'])
        }

    refine_delete_row = refine_insert_row

    def refine_update_row(self, row):
        updated_values = _get_updated_values(row['before_values'],
                                             row['after_values'])
        updated_values = self.remove_prefix_for_dict(updated_values)
        updated_values = self.remove_ignored_fields(updated_values)
        if not updated_values:
            return False
        return {
            'updated_values': updated_values,
            'values': self.remove_prefix_for_dict(row['after_values'])
        }

    def get_primary_key_values(self, row):
        vals = row['values']
        return dict([(pk, vals[pk]) for pk in self.key_fields])

    def make_data(self, event):
        """ Hàm này phục vụ các việc quyết định xem 1 binlog event có được bắn đi hay
        không? Nếu có thì dữ liệu bắn đi là gì?

        Các bước thực hiện như sau:

        1. Loại bỏ prefix của các field trên các rows dựa theo config.
        2. Loại bỏ các ignore field trong quá trình update.
        3. Loại bỏ các empty updated rows: các row được update nhưng không update field
           nào (sau khi đã loại bỏ bởi ignored fields).
        4. Trả về các not empty rows cùng `dbname` và `tbname`.

        Các row sau khi format sẽ có dạng:

        > Under construction

        :param pymysqlreplication.row_event.RowsEvent event: binlog event
        :return: False or (meta, rows)
        """
        raw_rows = event.rows
        event_action = _get_event_action(event)
        if event_action == 'insert':
            refine_func = self.refine_insert_row
        elif event_action == 'update':
            refine_func = self.refine_update_row
        elif event_action == 'delete':
            refine_func = self.refine_delete_row
        rows = []
        for r in raw_rows:
            r = refine_func(r)
            if not r:
                continue
            r['keys'] = self.get_primary_key_values(r)
            rows.append(r)

        if not rows:
            return False

        return ({
            'time': event.timestamp,
            'log_pos': event.packet.log_pos,
            'schema': self.schema,
            'table': self.table,
            'type': event_action,
        }, rows)


_event_data_makers = dict()


def get_event_data_maker(schema, table):
    global _event_data_makers
    schema_makers = _event_data_makers.setdefault(schema, {})
    try:
        data_maker = schema_makers[table]
    except KeyError:
        data_maker = TableAdapter(schema, table)
        schema_makers[table] = data_maker
    return data_maker


def make_event_data(event):
    """ This function make event data for a binlog event.

    The event data are rows and meta that contain data for rows and meta data
    the binlog event.

    Args:
        event (pymysqlreplication.row_event.RowsEvent): the binlog event

    Returns:
        tuple(rows, meta): The data that is made from event

        Where:

        **rows**


    Hàm này phục vụ các việc quyết định xem 1 binlog event có được bắn đi hay
    không? Nếu có thì dữ liệu bắn đi là gì?
    Các bước thực hiện như sau:

    1. Xuất phát từ event, đưa qua database mapping, table mapping để có `dbname` và
       `tbname` mới.
    2. Loại bỏ prefix của các field trên các rows dựa theo config.
    3. Loại bỏ các ignore field trong quá trình update.
    4. Loại bỏ các empty updated rows: các row được update nhưng không update field
       nào (sau khi đã loại bỏ bởi ignored fields).
    5. Trả về các not empty rows cùng `dbname` và `tbname`.

    Các row sau khi format sẽ có dạng:

    > Under construction

    :param pymysqlreplication.row_event.RowsEvent event: binlog event
    :return: False or (meta, rows)
    """
    # apply mappings
    schema = config.SCHEMA_NAME_MAPPINGS.get(event.schema, event.schema)
    table = config.TABLE_NAME_MAPPINGS.get(event.table, event.table)

    data_maker = get_event_data_maker(schema, table)
    return data_maker.make_data(event)



# coding=utf-8
import logging

__author__ = 'Tarzan'
_logger = logging.getLogger(__name__)


class BaseFilter(object):
    def __init__(self, conf):
        self.conf = conf

    def __call__(self, row):
        raise NotImplementedError()

    apply = __call__


_registered_filter_classes = {}


def register_filter(name, cls):
    global _registered_filter_classes
    _registered_filter_classes[name] = cls


def create_filter(name, conf):
    """ Create a filter with conf

    Returns:
        BaseFilter
    """
    return _registered_filter_classes[name](conf)


class ColumnPrefixFilter(BaseFilter):
    def __init__(self, conf):
        super(ColumnPrefixFilter, self).__init__(conf)
        self.prefix = conf
        self.lprefix = len(self.prefix)

    def __call__(self, row):
        if not self.lprefix:
            return True
        prefix = self.prefix
        lprefix = self.lprefix
        for values in row.values():
            keys = list(values)
            for k in keys:
                if k.startswith(prefix):
                    values[k[lprefix:]] = values.pop(k)
        return True


class HideColumnsFilter(BaseFilter):
    def __init__(self, conf):
        super(HideColumnsFilter, self).__init__(conf)
        self._columns_names = self.conf

    def __call__(self, row):
        if not self._columns_names:
            return True
        for values in row.values():
            for col in self._columns_names:
                values.pop(col, None)
        return True


class VerifyChangingFilter(BaseFilter):
    def __init__(self, conf):
        super(VerifyChangingFilter, self).__init__(conf)
        self._ignore_columns = set(self.conf)

    def __call__(self, row):
        if not self._ignore_columns:
            return True
        updated_values = row.get('updated_values', None)  # type: dict

        if updated_values is None:
            return True

        return not set(updated_values.keys()).issubset(self._ignore_columns)


register_filter('column_prefix', ColumnPrefixFilter)
register_filter('hide_columns', HideColumnsFilter)
register_filter('ignore_updated_columns', VerifyChangingFilter)


# coding=utf-8
import logging
import re

from mysqlbinlog2gpubsub import config
from mysqlbinlog2gpubsub.filters import factory


__author__ = 'Tarzan'
_logger = logging.getLogger(__name__)


def get_filters_for(table_name):
    """ Get list of filters for a table

    Args:
        table_name (str): table's name

    Returns:
        list[factory.BaseFilter]: list of filters
    """
    global _table_filters
    try:
        return _table_filters[table_name]
    except KeyError:
        filters = []
        for pattern, filters_conf in config.filters.items():
            if re.match('^%s$' % pattern, table_name):
                for fname, fconf in filters_conf.items():
                    filters.append(factory.create_filter(fname, fconf))
        _table_filters[table_name] = filters
        return filters


class Filters(object):
    """ Handle a filters chain """
    def __init__(self, filters_conf):
        self._filters_conf = filters_conf
        self._table_filters = {}
        self._filters_by_pattern = {}

    def _get_filters_for_pattern(self, pattern):
        try:
            return self._filters_by_pattern[pattern]
        except KeyError:
            filters = []
            for fname, fconf in self._filters_conf[pattern].items():
                filters.append(factory.create_filter(fname, fconf))
            self._filters_by_pattern[pattern] = filters
            return filters

    def get_filters_for_table(self, table_name):
        try:
            return self._table_filters[table_name]
        except KeyError:
            filters = []
            for pattern, filters_conf in self._filters_conf.items():
                if re.match('^%s$' % pattern, table_name):
                    filters.extend(self._get_filters_for_pattern(pattern))
            self._table_filters[table_name] = filters
            return filters

    def apply(self, rows, meta):
        """ Apply filters on rows and meta

        Args:
            rows (list[dict]): rows
            meta (dict): meta

        Every changes that happen on rows and meta will remain
        """
        table_name = '%s.%s' % (meta['schema'], meta['table'])
        for f in self.get_filters_for_table(table_name):
            for i in reversed(range(len(rows))):
                if not f(rows[i]):
                    del rows[i]
            if not rows:
                return False
        return True

    def yield_rows(self, rows, meta):
        """ Yield a list of rows that satisfy the filters

        Returns:
            list[dict]
        """
        table_name = '%s.%s' % (meta['schema'], meta['table'])
        filters = self.get_filters_for_table(table_name)
        for row in rows:
            if all([f(row) for f in filters]):
                yield row


# -*- coding: utf-8 -*-
import random
import gcloud._helpers
import yaml

__author__ = 'tarzan'


pubsub_project_name = None

mysql_settings = dict()

binlog_position_memory = (None, 2)

only_schemas = None
only_tables = None

schema_rename = dict()
table_rename = dict()

default_publisher_conf = {}

publishers = {}


def _import_from_dict(conf_dict):
    """ Import settings from a dictionary
    Args:
        conf_dict (dict): settings to be imported
    """
    globals().update(conf_dict)


def _import_from_yaml_file(yaml_file):
    with open(yaml_file) as f:
        conf = yaml.safe_load(f)
    return _import_from_dict(conf)


def _load_config_from_environment():
    import os

    conf_file = os.environ['BINLOG2GPUBSUB_CONF_FILE']
    _import_from_yaml_file(conf_file)


_load_config_from_environment()


# coding=utf-8
""" This module define class Publisher that controls the publishing progress"""
import json
import logging

import collections

import re

import gcloud.pubsub
import jose.jwk
import jose.jws

from mysqlbinlog2gpubsub import config
from mysqlbinlog2blinker import signals
from mysqlbinlog2gpubsub import filters as _filters

__author__ = 'Tarzan'
_logger = logging.getLogger(__name__)


def _json_default(o):
    import datetime
    if isinstance(o, (datetime.datetime, datetime.date)):
        return o.isoformat()
    if isinstance(o, (tuple, set)):
        return list(o)
    return o


_json_encoder = json.JSONEncoder(default=_json_default)
json_dumps = _json_encoder.encode


class Publisher(object):
    """ Control a publishing progress, from tables' binlog event and send to
    a pubsub topic
    """
    def __init__(self, name, topic_name, jwt, signals=None,
                 schema_rename=None, table_rename=None,
                 filters=None):
        """
        Args:
            name (str):
            topic_name (str):
            jwt (dict[str, str]): JWT configuration, including algorithm and key
            signals (dict[str, list]):
            schema_rename (dict[str, str]):
            table_rename (dict[str, dict[str, str]])
            filters (dict[str, mixed])
        """
        self.__name__ = name
        self.topic_name_pattern = topic_name
        self.signals_conf = signals

        self.schema_rename = schema_rename
        self._schema_new_names = dict()  # type: dict[str, str]
        self.table_rename = table_rename
        self._table_new_names = dict()  # type: dict[tuple(str, str), str]

        self.filters = _filters.Filters(filters or {})
        self.pubsub_client = gcloud.pubsub.Client(
            project=config.pubsub_project_name)
        self.pubsub_topics = dict()
        self.jwt_alg = jwt['alg']
        self.jwt_key = jose.jwk.RSA.importKey(jwt['key'])


    def _get_pubsub_topic(self, meta):
        """Get pubsub topic that corresponding to an event

        Returns:
            gcloud.pubsub.topic.Topic: the topic
        """
        topic_name = self.topic_name_pattern % meta
        try:
            return self.pubsub_topics[topic_name]
        except KeyError:
            topic = self.pubsub_client.topic(topic_name)
            if not topic.exists():
                topic.create()
            self.pubsub_topics[topic_name] = topic
            return topic

    def _bind_to_rows_signals(self):
        if self.signals_conf:
            for table_name, actions in self.signals_conf.items():
                actions = actions or ['insert', 'update', 'delete']
                if 'insert' in actions:
                    signals.rows_inserted.connect(self.on_rows_signal,
                                                  sender=table_name)
                if 'update' in actions:
                    signals.rows_updated.connect(self.on_rows_signal,
                                                 sender=table_name)
                if 'delete' in actions:
                    signals.rows_deleted.connect(self.on_rows_signal,
                                                 sender=table_name)
        else:
            signals.rows_inserted.connect(self.on_rows_signal)
            signals.rows_updated.connect(self.on_rows_signal)
            signals.rows_deleted.connect(self.on_rows_signal)

    def _unbind_to_rows_signals(self):
        if self.signals_conf:
            for table_name, actions in self.signals_conf.items():
                actions = actions or ['insert', 'update', 'delete']
                if 'insert' in actions:
                    signals.rows_inserted.disconnect(self.on_rows_signal,
                                                  sender=table_name)
                if 'update' in actions:
                    signals.rows_updated.disconnect(self.on_rows_signal,
                                                 sender=table_name)
                if 'delete' in actions:
                    signals.rows_deleted.disconnect(self.on_rows_signal,
                                                 sender=table_name)
        else:
            signals.rows_inserted.disconnect(self.on_rows_signal)
            signals.rows_updated.disconnect(self.on_rows_signal)
            signals.rows_deleted.disconnect(self.on_rows_signal)

    def start(self):
        """Start listening signals"""
        self._bind_to_rows_signals()

    def stop(self):
        """Stop listening signals"""
        self._unbind_to_rows_signals()

    def _rename_schema(self, schema):
        """ Rename a schema name and return new name

        Args:
            schema (str): schema nae

        Returns:
            str
        """
        try:
            return self._schema_new_names[schema]
        except KeyError:
            new_name = schema
            for pattern, repl in self.schema_rename.items():
                new_name, subs_made = re.subn(pattern, repl, schema)
                if subs_made:
                    break

            self._schema_new_names[schema] = new_name
            return new_name

    def _rename_table(self, schema, table):
        """ Rename a table and rereturn new name

        Args:
            schema (str): schema name (the new one)
            table (str): table name

        Returns:
            str
        """
        try:
            return self._table_new_names[(schema, table)]
        except KeyError:
            new_name = table
            _rules = self.table_rename.get(schema, {})
            for pattern, repl in _rules.items():
                new_name, subs_made = re.subn(pattern, repl, table)
                if subs_made:
                    break

            self._table_new_names[(schema, table)] = new_name
            return new_name

    def on_rows_signal(self, table_name, rows, meta):
        meta['schema'] = self._rename_schema(meta['schema'])
        meta['table'] = self._rename_table(meta['schema'], meta['table'])

        for row in self.filters.yield_rows(rows, meta):
            self.publish_row_to_pubsub(row, meta)

    def _make_jwt(self, row, meta):
        return jose.jws.sign(row, key=self.jwt_key, algorithm=self.jwt_alg)

    def publish_row_to_pubsub(self, row, meta):
        topic = self._get_pubsub_topic(meta=meta)  # type: gcloud.pubsub.Topic
        meta = {
            str(k): str(v)
            for k, v in meta.items()
        }

        # make row is json competible
        row = json.loads(json_dumps(row))
        _logger.info('Publishing %s.%s.%s#%s to %s' % (
            meta['schema'],
            meta['table'],
            meta['action'],
            row['keys'],
            topic.name
        ))

        msg_data = self._make_jwt(row, meta)
        topic.publish(msg_data, **meta)


_publishers = dict()  # type: dict[str, Publisher]


def init_publishers():
    """Read configuration and initialize publishers"""
    global _publishers  # type: dict

    def _merge_publisher_conf(conf, default):
        """ Merge default publisher configuration to a specific conf

        Args:
            conf (dict[str, dict]):
            default (dict[str, dict):

        Returns:

        """
        conf = conf.copy()
        conf.setdefault('schema_rename', {})
        conf.setdefault('table_rename', {})
        conf.setdefault('filters', {})
        conf.setdefault('jwt', {})
        conf['schema_rename'].update(default.get('schema_rename', {}))
        conf['table_rename'].update(default.get('table_rename', {}))
        conf['filters'].update(default.get('filters', {}))
        conf['jwt'].update(default.get('jwt', {}))
        return conf

    default_publisher_conf = config.default_publisher_conf

    for p_name, p_conf in config.publishers.items():
        _publishers[p_name] = Publisher(
            p_name,
            **_merge_publisher_conf(p_conf, default_publisher_conf))

    for p in _publishers.values():
        p.start()


# -*- coding: utf-8 -*-
"""Project metadata
Information describing the project.
"""
import logging

__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


# The package name, which is also the "UNIX name" for the project.
package = 'mysqlbinlog2gpubsub'
project = "MySQL binlog to Google Cloud Pub/Sub"
project_no_spaces = project.replace(' ', '')
version = '1.0.7'
description = 'Send MySQL binlog event to Google Cloud Pub/Sub'
authors = ['Tarzan', ]
authors_string = ', '.join(authors)
emails = ['hoc3010@gmail.com', ]
license = 'MIT'
copyright = '2016 ' + authors_string
url = 'https://github.com/tarzanjw/python-mysql-binlog-pubsub'



# -*- coding: utf-8 -*-
import functools
import json
import logging
import gcloud.pubsub
from mysqlbinlog2gpubsub import config

__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


def _json_default(o):
    import datetime
    if isinstance(o, (datetime.datetime, datetime.date)):
        return o.isoformat()
    if isinstance(o, (tuple, set)):
        return list(o)
    return o


_json_encoder = json.JSONEncoder(default=_json_default)
json_dumps = _json_encoder.encode


pubsub_client = gcloud.pubsub.Client(project=config.PUBSUB_PROJECT_NAME)


def publish_rows(topic_name, rows, meta):
    topic = pubsub_client.topic(topic_name)

    rows_message = json_dumps(rows).encode('utf-8')
    attrs = {
        k: str(v)
        for k, v in meta.items()
    }

    _logger.debug('Publish %d rows to topic "%s"' % (len(rows), topic_name))
    topic.publish(rows_message, **attrs)


# -*- coding: utf-8 -*-
import logging

import mysqlbinlog2blinker
import mysqlbinlog2blinker.signals


__author__ = 'tarzan'
_logger = logging.getLogger(__name__)


def start_publishing():
    from mysqlbinlog2gpubsub import config
    from mysqlbinlog2gpubsub import publishers

    publishers.init_publishers()
    mysqlbinlog2blinker.start_replication(
        mysql_settings=config.mysql_settings,
        binlog_pos_memory=config.binlog_position_memory,
        only_schemas=config.only_schemas,
        only_tables=config.only_tables,
    )